import { Client, Intents } from "discord.js";
import chalk from "chalk";
import dotenv from "dotenv";
dotenv.config();
import { sLogger as Logger } from "../logger/logger";
import { Version as FrameworkVersion } from "../config";
import { VoiceControl, VoiceValidateResult } from "../voice";
import { checkPrefix } from "../utils/string";
export class SBotClient {
    constructor(options) {
        this.props = {};
        this.corgiSwiftQueue = [];
        this.client = new Client({
            intents: [
                Intents.FLAGS.GUILDS,
                Intents.FLAGS.GUILD_MESSAGES,
                Intents.FLAGS.GUILD_MESSAGE_REACTIONS,
                Intents.FLAGS.GUILD_VOICE_STATES,
                Intents.FLAGS.DIRECT_MESSAGES,
                Intents.FLAGS.DIRECT_MESSAGE_REACTIONS,
            ],
            partials: ["CHANNEL"],
        });
        console.log(chalk.cyan("Starting S-Bot Framework ") +
            chalk.magenta(FrameworkVersion) +
            "✨✨");
        // * Initialize Utility Classes
        this.utility = {
            response: [],
        };
        const { token = process.env.DISCORD_TOKEN, activityRefreshInterval = 120, logLocation, } = options ?? {};
        Logger.startFile(logLocation);
        this.client.on("ready", () => {
            Logger.log(`=====> Logged in! Bot Mounted on ${this.client.user?.tag} <=====`, "SPECIAL", false);
            this.setSelfActivity();
        });
        this.client.login(token);
        setInterval((() => {
            this.setSelfActivity();
        }).bind(this), activityRefreshInterval * 1000);
        this.client.on("messageCreate", this.response.bind(this));
    }
    get user() {
        return this.client.user;
    }
    useComputedActivity(activity) {
        const { type, name = "" } = activity;
        this.props.computedActivity = activity;
        this.client.user?.setActivity(name, { type, name });
    }
    useActivities(loader) {
        this.utility.activityLoader = loader;
    }
    setSelfActivity() {
        const loader = this.utility.activityLoader;
        if (!this.props.computedActivity && !loader?.length)
            return;
        const selectedIndex = Math.floor(Math.random() *
            (Number(!!loader?.length) +
                Number(!!this.utility.activityLoader)));
        const selectedActivity = loader?.getData()[selectedIndex] ?? this.props.computedActivity;
        if (selectedActivity) {
            const { type, name = "" } = selectedActivity;
            this.client.user?.setActivity(name, { type, name });
        }
    }
    useResponse(response) {
        response.setClient(this);
        this.utility.response.push(response);
    }
    response(msg) {
        if (msg.author.id == this.client.user?.id)
            return;
        if (msg.author.bot)
            return;
        Logger.log(`Recieved Message from ${msg.author.tag}: ${msg.content}`);
        if (this.voiceOptions?.jutsu == "SOnDemand") {
            if (checkPrefix(msg.content, this.voiceOptions.prefix.join)) {
                this.sodJoin(msg);
                return;
            }
            if (checkPrefix(msg.content, this.voiceOptions.prefix.leave)) {
                this.sodLeave(msg);
                return;
            }
        }
        for (const response of this.utility.response) {
            if (response.isTrigger(msg.content)) {
                const reply = response.getReply();
                try {
                    if (reply.react)
                        msg.react(reply.react);
                    if (reply.reply)
                        msg.reply(reply.message);
                    else
                        msg.channel.send(reply.message);
                    if (reply.audio)
                        this.ttsJutsu(msg, reply.message);
                    const refInd = reply.refIndex;
                    Logger.log(`Replied ${msg.author.tag} with ${reply.message}${refInd ? ` (${refInd})` : ""}`);
                }
                catch (err) {
                    Logger.log(`Error while responding to message : ${err}`, "ERROR");
                }
                return;
            }
        }
    }
    useConsole(console) {
        this.utility.console = console;
    }
    useVoice(options) {
        this.voiceOptions = options;
    }
    async ttsJutsu(msg, content) {
        if (!this.voiceOptions?.jutsu)
            return;
        if (this.voiceOptions.jutsu == "SOnDemand") {
            if (!this.voiceCtrl?.isSameChannel(msg.member?.voice.channel)) {
                return;
            }
            if (this.voiceCtrl) {
                this.voiceCtrl.speak(content);
                return;
            }
        }
        if (this.voiceOptions.jutsu == "CorgiSwift") {
            this.corgiSwiftJutsu(msg, content);
            return;
        }
    }
    // TODO Check if already joined
    async sodJoin(msg) {
        const voiceOptions = this.voiceOptions;
        try {
            this.voiceCtrl = new VoiceControl(msg, (() => {
                this.voiceCtrl = undefined;
            }).bind(this));
            await this.voiceCtrl.waitTillReady("SOD Mode");
            if (voiceOptions.onJoin) {
                await this.voiceCtrl.speak(voiceOptions.onJoin);
            }
        }
        catch (err) {
            if (voiceOptions.fallback?.join_fail?.message) {
                const message = (() => {
                    const messages = voiceOptions.fallback.join_fail.message;
                    switch (err) {
                        case VoiceValidateResult.NO_CHANNEL:
                            return messages.no_channel;
                        case VoiceValidateResult.STAGE_CHANNEL:
                            return messages.stage_channel;
                        case VoiceValidateResult.NOT_JOINABLE:
                            return messages.not_joinable;
                        default:
                            return messages.internal;
                    }
                })();
                if (!message)
                    return;
                if (voiceOptions.fallback.join_fail.reply)
                    msg.reply(message);
                else
                    msg.channel.send(message);
            }
        }
    }
    async sodLeave(msg) {
        const voiceOptions = this.voiceOptions;
        if (!this.voiceCtrl?.isSameChannel(msg.member?.voice.channel) &&
            voiceOptions.rules?.onsite_leave) {
            const message = voiceOptions.fallback?.onsite_leave?.message;
            if (message) {
                if (voiceOptions.fallback?.onsite_leave?.reply)
                    msg.reply(message);
                else
                    msg.channel.send(message);
            }
            return;
        }
        this.voiceCtrl?.destruct();
    }
    async corgiSwiftJutsu(msg, content) {
        if (VoiceControl.validateUser(msg))
            return;
        const notRunning = !this.corgiSwiftQueue.length;
        this.corgiSwiftQueue.push({ msg, content });
        if (notRunning)
            this.corgiSwiftClearQueue();
    }
    async corgiSwiftClearQueue() {
        if (!this.corgiSwiftQueue.length)
            return;
        const { msg } = this.corgiSwiftQueue[0];
        const targetVoiceChannel = msg.member?.voice.channel;
        if (!targetVoiceChannel?.joinable) {
            // * Users no longer in VC or VC is not joinable
            this.corgiSwiftQueue.shift();
            this.corgiSwiftClearQueue();
            return;
        }
        try {
            this.voiceCtrl = new VoiceControl(msg, (() => {
                this.voiceCtrl = undefined;
            }).bind(this));
            await this.voiceCtrl.waitTillReady("CorgiSwift術 Mode");
            while (this.corgiSwiftQueue.length) {
                const words = this.corgiSwiftQueue[0];
                if (this.voiceCtrl.isSameChannel(words.msg.member?.voice.channel)) {
                    await this.voiceCtrl.speak(words.content);
                    this.corgiSwiftQueue.shift();
                }
                else {
                    this.voiceCtrl.destruct();
                    this.corgiSwiftClearQueue();
                    return;
                }
            }
            this.voiceCtrl?.destruct();
        }
        catch (err) {
            Logger.log(`CorgiSwift術 Failed: ${err}`);
            this.voiceCtrl?.destruct();
            this.corgiSwiftQueue.shift();
            this.corgiSwiftClearQueue();
        }
    }
}
